import time

class zglFadeObject:
	def __init__(self):
		self.alpha = 1.0
		self.fading_in = False
		self.fading_out = False
		self.fade_start_alpha = 1.0
		self.last_step_time = 0
		self.fade_in_delay = 0.0
		self.fade_in_time = 0.250		# in seconds
		self.fade_out_delay = 0.0
		self.fade_out_time = 0.250

	def fadeIn(self, t=time.time()):
		self.fading_in = True
		self.fading_out = False
		self.last_step_time = t + self.fade_in_delay

	def fadeOut(self, t=time.time()):
		self.fading_in = False
		self.fading_out = True
		self.last_step_time = t + self.fade_out_delay

	# time should be passed in to ensure the animation on each frame
	# are synchronized in terms of time
	def stepFade(self, t):
		if self.fading_in and t > self.last_step_time:
			d_time = t - self.last_step_time
			self.alpha += d_time/self.fade_in_time
			if self.alpha > 1.0:
				self.alpha = 1.0
				self.fading_in = False
			self.last_step_time = t
		elif self.fading_out and t > self.last_step_time:
			d_time = t - self.last_step_time
			self.alpha -= d_time/self.fade_out_time
			if self.alpha < 0.0:
				self.alpha = 0.0
				self.fading_out = True
			self.last_step_time = t



import logging
import math, sys, os, time

import wx
import wx.glcanvas 

import Event as EventModule

class wxFrame(wx.Frame):
    def __init__(self, *args, **kwds):
        wx.Frame.__init__(self, *args, **kwds)
        self.canvas = None

        return



    def bindEvents(self):
        self.Bind(wx.EVT_SHOW, self.OnShowFrame)
        self.Bind(wx.EVT_SIZE, self.OnResize)
        return
    
    def OnResize(self, event):
        self.Skip()
        return
    
    def OnShowFrame(self, event):
        self.canvas.displayFunc()
        self.Skip()
        return

    
    
    # END wxFrame
    pass





import zgl_graphdrawer.canvas.gl_canvas as CanvasModule

import wx
import wx.glcanvas 




class Canvas(wx.glcanvas.GLCanvas, CanvasModule.Canvas):

    def __init__(self, parent, *args, **kwds):

        attribList = (
            wx.glcanvas.WX_GL_DOUBLEBUFFER,
            wx.glcanvas.WX_GL_RGBA, 0
        )
        wx.glcanvas.GLCanvas.__init__(
            self, parent, -1, attribList = attribList,
            *args, **kwds)

        CanvasModule.Canvas.__init__(self, parent, *args, **kwds)

        return

    def initializeTimer(self):

        self.TIMER_ID = 100  # pick a number

        # message will be sent to the parent
        self.timer = wx.Timer(self.GetParent(), self.TIMER_ID)  
        self.timer.Start(100)  # x100 milliseconds

        # TODO:
        # figure out how to have multiple listeners for the time event
        # currently only the CanvasEventHandler is handling it

        return


    def forceRedraw(self):
        dc = wx.PaintDC(self)
        self.Render(dc)
        return

    def Render(self, dc):
        self.SetCurrent()

        self.displayFunc()

        self.SwapBuffers()
        return

    def children(self):
        for node in self.nodes:
            yield node
        for edge in self.edges:
            yield edge
        raise StopIteration

    # END class Canvas
    pass




import logging
import sys

import zgl_graphdrawer.Edge as EdgeModule
import zgl_graphdrawer.Node as NodeModule


class ContextManager(object):

    def __init__(self):

        # initialize the dictionry that holds
        # the locations of the commands that may be called
        self._commands = {}
        return

    def app(self, value=None):
        if value is not None:
            self._app = value
        return self._app


    def initializeCanvas(self, canvas):

        self.canvas = canvas
        return

    def initializeEventHandlers(self):

        self.canvas.initializeEventHandlers(contextManager=self)

        pass


    def resourcePath(self, value=None):
        """
        this sets the path from which the resources should be loaded
        """

        if value is not None:
            self._resourcePath = value
        if not hasattr(self, '_resourcePath'):
            self._resourcePath = '.'
        return self._resourcePath

    
    def commandPath(self, key, value=None):
        """
        this sets the path of a command
        """

        if value is not None:
            self._commands[key] = value
        return self._commands[key]




    def selectionChanged(self, oldSelection, newSelection):
        for object in oldSelection:
            if object not in newSelection and object.isSelectable():
                # The object was removed from selection
                object.isSelected(False)

                self.canvas.nodePolicy().updateNode(object)

        for object in newSelection:
            if object not in oldSelection and object.isSelectable():
                # The object was added to the selection
                object.isSelected(True)

                # This code assumes selection is a node
                # which for now is true but may not be in the future
                self.canvas.nodes.remove(object)
                self.canvas.nodes.append(object)
                self.canvas.nodePolicy().updateNode(object)

        return


    def objectsMoved(self, objects):
        for edge in self.edges.values():
            if edge.inputNode in objects or \
               edge.outputNode in objects:
                edge.setupPrimitives(self.canvas.edgePolicy())
                pass
            pass
        return


    def draw(self):
        """
        TODO: move out of context manager class
        """
        try:
            self.canvas.draw()
        except Exception, e:
            logging.error(e)
            sys.exit(1)
        return


    def canConnect(self, port1Data, port2Data):
        print "need to determine if can actually connect ports"
        return True


    def connect(self, event, port1Data, port2Data):
        print "should connect %s -> %s" % (
            port1Data[0].getInputPort(port1Data[1]),
            port2Data[0].getOutputPort(port2Data[1])
        )
        return


    # END class ContextManager
    pass


class GraphDrawerManager(ContextManager):

    MOUSEMODE_SELECT = 0
    MOUSEMODE_CANVAS_PAN = 1
    MOUSEMODE_CANVAS_ZOOM = 2
    MOUSEMODE_DRAG_OBJECT = 3

    MOUSEMODE_HOTKEYS = {
        ' ':MOUSEMODE_CANVAS_PAN,
        'z':MOUSEMODE_CANVAS_ZOOM
    }



    def initialize(self, canvas=None):

        if canvas is None:
            import zgl_graphdrawer.canvas.zgl_canvas as CanvasModule
            canvas = CanvasModule.Canvas()

        # Selection Service
        self.selection = NodeSelection(self)

        self.initializeCanvas(canvas)

        pass



    def updateFrame(self, frame):
        self.canvas.updateFrame(frame)




    def step(self):
        """ called every frame. For animation purposes. """
        return


    def keyDown(self, key):
        mouseHandler = self.canvas.activeMouseHandler()
        if not mouseHandler.isActive():
            self.canvas.activeMouseMode = GraphDrawerManager.MOUSEMODE_HOTKEYS.get(key, None)
            pass
        return


    def keyUp(self, key):
        """
        default to selection
        """
        self.canvas.activeMouseMode = GraphDrawerManager.MOUSEMODE_SELECT
        return


    def mouseMoved(self, x, y):
        self.canvas.mouseMoved(x, y)


    def mouseButtonDown(self, button, x, y):


        result = self.canvas.mouseButtonDown(button, x, y)

        if result is not None and result != True:
            # selection clicked on an object
            self.canvas.mouseHandlers[GraphDrawerManager.MOUSEMODE_DRAG_OBJECT] = \
                ObjectDragMoveMouseHandler(self.selection.selection, self.canvas, self)

            # need to cache the previous mouse mode
            self.prevMouseMode = self.canvas.activeMouseMode
            self.canvas.activeMouseMode = GraphDrawerManager.MOUSEMODE_DRAG_OBJECT
            self.canvas.mouseButtonDown(button, x, y)

        return

    def mouseDragged(self, x, y):
        self.canvas.mouseDragged(x, y)

    def mouseButtonUp(self, button, x, y):

        self.canvas.mouseButtonUp(button, x, y)

        # if the mousedown had changed the mouse mode
        # we restore it here
        if self.prevMouseMode is not None:
            self.canvas.activeMouseMode = self.prevMouseMode
            self.prevMouseMode = None
        return

    def setDataContext(self, nodes, edges):
        """
        This is only used by the zglCanvas side
        """
        self.nodes = {}
        self.edges = {}

        for nodeData in nodes:
            uiNode = NodeModule.Node(nodeData)
            self.nodes[nodeData] = uiNode

        for edgeData in edges:
            uiEdge = EdgeModule.Edge(edgeData)
            if edgeData.inputNode() is not None:
                uiEdge.inputNode(self.nodes[edgeData.inputNode()])


            if edgeData.outputNode() is not None:

                uiEdge.outputNode(self.nodes[edgeData.outputNode()])

            self.edges[edgeData] = uiEdge
            pass

        self.canvas.resetDrawables()
        map(self.canvas.addDrawable, self.nodes.itervalues())
        map(self.canvas.addDrawable, self.edges.itervalues())

        logging.debug("GraphDrawer setup %d nodes and %d edges" % 
                      (len(self.nodes), len(self.edges)))
        return


    # END class GraphDrawerManager
    pass



class NodeSelection:
    def __init__(self, delegate):
        self.selection = []
        self.delegate = delegate

    def count(self):
        return len(self.selection)

    def setSelection(self, newSelection):
        oldSelection = self.selection
        self.selection = newSelection
        self.delegate.selectionChanged(oldSelection, newSelection)

        return



import logging

from Responder import *
	
import zgl.zglPrimitives as PrimitivesModule

# An Edge defines a connection between 
#			
class Edge(Responder):
	def __init__(self, edgeData):
		Responder.__init__(self)
		self.edgeData = edgeData
		self._inputNode = None
		self._outputNode = None
		
		return
	
	def sourceNode(self, *args, **kwds):
		return self.inputNode(*args, **kwds)

	def targetNode(self, *args, **kwds):
		return self.outputNode(*args, **kwds)

	def outputNode(self, value=None):
		if value is not None:
			self._outputNode = value
		return self._outputNode

	def inputNode(self, value=None):
		if value is not None:
			self._inputNode = value
		return self._inputNode


	def inputPort(self):
		if self.inputNode() is None:
			raise AttributeError('no attribute "inputNode" found')
		return self.inputNode().getOutputPort(self.edgeData.inputPort)
	
	
	def outputPort(self):
		if self.outputNode() is None:
			raise AttributeError('no attribute "outputNode" found')
		return self.outputNode().getInputPort(self.edgeData.outputPort)

	def setupPrimitives(self, edgePolicy):
		self.children = []
					
		try:
			inputPort = self.inputPort()
			outputPort = self.outputPort()
			
			point1 = (self.inputNode().x + inputPort.x + inputPort.width/2, 
				  self.inputNode().y + inputPort.y + inputPort.height/2)
			point2 = (self.outputNode().x + outputPort.x + outputPort.width/2, 
				  self.outputNode().y + outputPort.y + outputPort.height/2)
		except AttributeError, e:
			logging.error(e)
			return
		
		line = PrimitivesModule.zglLine()
		line.colour = edgePolicy._lineColour
		line.points = edgePolicy.createPath(point1, point2, 30)
		self.children.append(line)
		
		edgePolicy.updateEdge(self)
		return

	
	def draw(self):
		for child in self.children:
			child.draw()


	# Override hit detection
	# Do not allow edges to be selected for now
	def isInside(self, x, y):
		return False


	def intersectsRect(self, rect):
		return False
		

	# END class Edge
	pass



import logging
import os
import re
import subprocess
import tempfile

import numpy


class LayoutPolicy(object):
    def __init__(self, contextManager):
        self._manager = contextManager
        return

    def layoutNodes(self, nodes, edges):
        raise NotImplementedError('need to implement %s.layoutNodes' % self.__class)

    # END class LayoutPolicy
    pass 


class GraphVizLayoutPolicy(LayoutPolicy):

    def temporaryFilePath(self, value=None):
        if value is not None:
            self._tmpFile = None
            
        if not hasattr(self, '_tmpFile'):
            self._tmpFile = os.sep.join(
                [tempfile.gettempdir(), 'temp_dotFile.dot'])
            
        return self._tmpFile


    def generateDotFile(self, path, nodes, edges):

        dotFile = "digraph G {\n"

        #for node in nodes:
        #	dotFile += "\t\"%s\"\n" % node.nodeData.name()

        for edge in edges:
            dotFile += "\t\"%s\" -> \"%s\"\n" % (edge.inputNode().nodeData.name(), edge.outputNode().nodeData.name())
        dotFile += "}"

        logging.debug(dotFile)

        fp = open(path, 'w')
        fp.write(dotFile)
        fp.close()

        return


    def layoutNodes(self, nodes, edges):

        tmpFile = self.temporaryFilePath()
        self.generateDotFile(tmpFile, nodes, edges)

        # use the subprocess module to execute
        layoutFile = '/tmp/pomset.layout'
        command = [self._manager.commandPath('dot'),
                   tmpFile, '-o%s' % layoutFile]

        try:
            ret = subprocess.call(command)
        except OSError:
            pass

        boundingbox = ""

        with open(layoutFile, 'r') as f:
            for line in f.readlines():
                logging.debug(line)

                match = re.search('^\s*graph \[(.*)\];$', line)
                if match is not None:
                    attributeString = match.group(1)
                    logging.debug("Graph has attributes: %s" % match.group(1))
                    attributes = self.getAttributesDict(attributeString)
                    if attributes.has_key('bb'):
                        bb = map(float, attributes['bb'].split(','))
                    continue;

                foundNode = False
                for node in nodes:
                    nodeName = "%s" % (node.nodeData.name())
                    match = re.search('^\s*"{0,1}'+nodeName+'"{0,1}\s*\[(.*)\];$', line)
                    if match is not None:
                        attributeString = match.group(1)
                        attributes = self.getAttributesDict(attributeString)
                        if attributes.has_key('pos'):

                            position = 1.5 * numpy.matrix(
                                ' '.join(attributes['pos'].split(',')))
                            logging.debug('position >> %s' % position)
                            node.setPosition(position[0,0], position[0,1])
                            pass
                        if attributes.has_key('width'):
                            (node.width) = float(attributes['width']) * 80
                        if attributes.has_key('height'):
                            (node.height) = float(attributes['height']) * 80
                        foundNode = True;
                    if foundNode:
                        break

        return (bb[2],bb[3])

    
    def getAttributesDict(self, attributeString):
        dict = {}
        matches = re.findall('(\w+)="(.*?)"',attributeString)
        if matches is not None:
            for match in matches:
                attributeName = match[0]
                attributeValue = match[1]
                dict[attributeName] = attributeValue
        return dict



import os, string
from OpenGL.GL import *
from OpenGL.GLUT import *
from zglUtils import *

# General shader class
class zglShaders:
	use_GL_functions = isFunctionAvailable("glCreateProgram");
	use_ARB_functions = False;
	
	def __init__(self):
		self.program_id = -1
		
		self.program_id = glCreateProgram()

	
	def useVertexShader(self, shader_file):
		file = open(shader_file, 'r')
		shader_source = file.read()
		
		vertex_shader_id = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(vertex_shader_id, shader_source);
		glCompileShader(vertex_shader_id);
#		self.vertex_shader_id = vertex_shader_id
		glAttachShader(self.program_id, vertex_shader_id)
	
	
	def useFragmentShader(self, shader_file):
		file = open(shader_file, 'r')
		shader_source = file.read()
		
		fragment_shader_id = glCreateShader(GL_FRAGMENT_SHADER);
		glShaderSource(fragment_shader_id, shader_source);
		glCompileShader(fragment_shader_id);
#		self.fragment_shader_id = fragment_shader_id
		glAttachShader(self.program_id, fragment_shader_id)
		
	
		

class zglShaderManager:
	vertex_shaders = { }
	fragment_shaders = { }
	
	def getVertexShader(self, vertex_shader_file):
		file = open(vertex_shader_file, 'r')
		shader_source = file.read()
		
		vertex_shader_id = glCreateShader(GL_VERTEX_SHADER);
		glShaderSource(vertex_shader_id, shader_source);
		glCompileShader(vertex_shader_id);
		
		vertex_shaders[vertex_shader_id] = vertex_shader_id



from zgl.zglUtils import *
from zgl.zglText import *

from MouseHandlers import MouseHandler



class Responder(object):
	def __init__(self):
		self.is_mouseover = False
		self.isSelected(False)
		return

	def visible(self, value=None):
		if value is not None:
			self._visible = value
		if not hasattr(self, '_visible'):
			self._visible = True
		return self._visible

	def isSelected(self, value=None):
		if not self.isSelectable():
			return False
		if value is not None:
			self._isSelected = value
		return self._isSelected
		
	def isSelectable(self):
		return False

	def isClickable(self):
		return False
	
	def overlaps(self, point):
		return False
		
	def mouseMoved(self, x, y):
		if self.isInside(x,y):
			self.is_mouseover = True
		return False
	
	def mouseButtonDown(self, button, x, y):
		return False
	
	def mouseButtonUp(self, button, x, y):
		return False
	
	def mouseDragged(self, x, y):
		return False
	



#!/usr/bin/env python
import sys, os
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *



DEFAULT_WIDTH = 800
DEFAULT_HEIGHT = 600

class zglApplication(object):

	def __init__(self, 
		     width=DEFAULT_WIDTH, 
		     height=DEFAULT_HEIGHT, 
		     app_name=""):
		"""initialize"""

		self.width = width
		self.height = height
		self.app_name = app_name
		self.path = os.getcwd()
		self.initGL()
		self.initLighting()
		self.initViewport()

		
	def initGL(self):
		glutInit(sys.argv)
		# reset the path here because GLUT will change paths to its own path
		os.chdir(self.path)
	
		glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)
		glutInitWindowSize(self.width, self.height)
		glutInitWindowPosition(0, 0)
	
		# Okay, like the C version we retain the window id to use when closing, but for those of you new
		# to Python (like myself), remember this assignment would make the variable local and not global
		# if it weren't for the global declaration at the start of main.
		self.window = glutCreateWindow(self.app_name)
		
		glClearDepth (1.0)
		glEnable (GL_DEPTH_TEST)
		glClearColor (0.0, 0.0, 0.0, 0.0)
		glShadeModel (GL_SMOOTH)
		glMatrixMode (GL_PROJECTION)
		
	
	def initLighting(self):
		# initialize lighting */
		lightOnePosition = (40.0, 40, 100.0, 0.0)
		lightOneColor = (0.99, 0.99, 0.99, 1.0) 
		lightTwoPosition = (-40.0, 40, 100.0, 0.0)
		lightTwoColor = (0.99, 0.99, 0.99, 1.0) 

		glLightfv (GL_LIGHT0, GL_POSITION, lightOnePosition)
		glLightfv (GL_LIGHT0, GL_DIFFUSE, lightOneColor)
		glEnable (GL_LIGHT0)
		glLightfv (GL_LIGHT1, GL_POSITION, lightTwoPosition)
		glLightfv (GL_LIGHT1, GL_DIFFUSE, lightTwoColor)
		glEnable (GL_LIGHT1)
		glEnable (GL_LIGHTING)
		glColorMaterial (GL_FRONT_AND_BACK, GL_DIFFUSE)
		glEnable (GL_COLOR_MATERIAL)
	
	def initViewport(self):
		# roughly, measured in centimeters */
		glViewport(0, 0, self.width, self.height)		# Reset The Current Viewport And Perspective Transformation
		self.updateProjection()
		
	def mousePassiveFunc(self, x, y):
		self.last_mouse_x = x
		self.last_mouse_y = y
		glutPostRedisplay()		
	
	def mouseFunc(self, button, state, x, y):
		""" mouseFunc """
	
	def mouseMotionFunc(self, x, y):
		self.last_mouse_x = x
		self.last_mouse_y = y
		glutPostRedisplay()
		
	def keyboardFunc(self, key, x, y):
		# If escape is pressed, kill everything.
		if key == ESCAPE:
			sys.exit()
			
	def keyboardUpFunc(self, key, x, y):
		pass
			
	def specialFunc(self, *args):
		pass
			
	def displayFunc(self):
		self.updateProjection()
		self.updateLighting()
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
		glDrawBuffer(GL_BACK)
		glutSwapBuffers()
	
	
	def reshapeFunc(self, width, height):
#		print "Reshaping Window "+str(width)+","+str(height)
		if width == 0:
			width = 1
		self.width = width
		self.height = height
		glViewport(0, 0, self.width, self.height)		# Reset The Current Viewport And Perspective Transformation
		self.updateProjection()
	
	
	def idleFunc(self):
		glutPostRedisplay()
		return

	def updateProjection(self):
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		gluPerspective(45.0, 1.0*self.width/self.height, 0.1, 100.0)
		glMatrixMode(GL_MODELVIEW)

	#----------------------------------------------
	#	Make a 2D Ortho view for this window size.
	#	Used for 2D drawing (UI, for example)
	def updateProjection2DOrtho(self):
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		gluOrtho2D(0.0, self.width, 0.0, self.height)
		glMatrixMode(GL_MODELVIEW)
		glLoadIdentity()
		
	
	def updateLighting(self):
		""" update lighting """

	
	def run(self):
		glutDisplayFunc(self.displayFunc)
		glutReshapeFunc(self.reshapeFunc)
		glutIdleFunc(self.idleFunc)
		glutMotionFunc(self.mouseMotionFunc)
		glutKeyboardFunc(self.keyboardFunc)
		glutKeyboardUpFunc(self.keyboardUpFunc)
		glutSpecialFunc(self.specialFunc)
		glutPassiveMotionFunc(self.mousePassiveFunc)
		glutMouseFunc(self.mouseFunc)
		glutMotionFunc(self.mouseMotionFunc)
		glutMainLoop()

		
	# END class zglApplication
	pass


if __name__ == "__main__":
	# execution stuff
	app = zglApplication(width, height, "PyOpenGLApplication")
	app.run()





from OpenGL.GL import *
from OpenGL.GLUT import *


class zglConsole:
	def __init__(self):
		"""init"""
		self.lines = []		# array to store the lines
		self.line_spacing = 18
		self.font = GLUT_BITMAP_HELVETICA_10
		
	def setFont(self, _font):
		self.font = _font
		
	def draw(self):
		""" """
		xx = x
		yy = self.y
		for line in lines:
			zglUtils.drawBitmapText(line, xx, yy, self.font, self.line_spacing)
			
			yy = yy + 18


import logging


import numpy

import zgl.zglPrimitives as PrimitivesModule
import zgl.zglText as TextModule

import zgl_graphdrawer.Menu as MenuModule
from zgl_graphdrawer.NodePolicy import *
import zgl_graphdrawer.Port as PortModule
from zgl_graphdrawer.Responder import *


class Node(Responder):

    PART_BORDER = "node border"
    PART_BACKGROUND = "node background"
    PART_LABEL = "node label"


    @staticmethod
    def hasCustomDragReleaseCallback():
        return False

    @staticmethod
    def OnDrag(event, canvas=None,
               objects=None, 
               basePosition=None,
               resetCallback=None,
               shouldResetCallbackFunction=None):

        if objects is None:
            raise NotImplementedError('should not happen')
        
        if shouldResetCallbackFunction(event):
            resetCallback()

        originalCanvasMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
            (basePosition.x, basePosition.y))

        eventPosition = event.GetPosition()
        eventCanvasMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
            (eventPosition.x, eventPosition.y))

        originalWorldMatrix = canvas.getWorldCoordinatesFromCanvasCoordinates(originalCanvasMatrix)
        eventWorldMatrix = canvas.getWorldCoordinatesFromCanvasCoordinates(eventCanvasMatrix)
        worldDeltaMatrix = eventWorldMatrix - originalWorldMatrix

        delta = (worldDeltaMatrix[3,0], worldDeltaMatrix[3,1])

        for object in objects:
            object.setPosition(object.initialX + delta[0],
                               object.initialY + delta[1])
            pass

        edgesToProcess = filter(
            lambda x: x.inputNode() in objects or x.outputNode() in objects,
            canvas.edges
        )
        map(lambda x: x.setupPrimitives(canvas.edgePolicy()), edgesToProcess)

        return



    def __init__(self, nodeData):
        
        Responder.__init__(self)
        self.x = 0
        self.y = 0
        self.width = 0
        self.height = 0
        self.inputPorts = {}
        self.outputPorts = {}
        self.parts = {
            "inputPorts":self.inputPorts,
            "outputPorts":self.outputPorts
            }

        
        self.nodeData = nodeData
        
        self.shouldUpdateDisplay(True)
        self.shouldUpdateImage(False)
        self.shouldUpdateColours(True)

        return


    def contextManager(self, value=None):
        if value is not None:
            self._contextManager = value
        if not hasattr(self, '_contextManager'):
            self._contextManager = None
        return self._contextManager
 
    
    def canvas(self):
        return self.contextManager.canvas

    
    def shouldUpdateDisplay(self, value=None):
        if value is not None:
            self._shouldUpdateDisplay = value
        if not hasattr(self, "_shouldUpdateDisplay"):
            self._shouldUpdateDisplay = False
        return self._shouldUpdateDisplay

    def shouldUpdateImage(self, value=None):
        if value is not None:
            self._shouldUpdateImage = value
        if not hasattr(self, "_shouldUpdateImage"):
            self._shouldUpdateImage = False
        return self._shouldUpdateImage


    def shouldUpdateColours(self, value=None):
        if value is not None:
            self._shouldUpdateColours = value
        if not hasattr(self, "_shouldUpdateColours"):
            self._shouldUpdateColours = False
        return self._shouldUpdateColours


    def isInside(self, x, y):
        if x < self.x or x > self.x+self.width:
            return False
        if y < self.y or y > self.y+self.height:
            return False
        return True

    def overlaps(self, point):
        if point.x < self.x:
            return False
        if point.x > self.x+self.width:
            return False
        if point.y < self.y:
            return False
        if point.y > self.y+self.height:
            return False
        return True


    def intersectsRect(self, rect):
        if (rect[0]+rect[2] > self.x) and \
           (rect[0] < self.x+self.width) and \
           (rect[1]+rect[3] > self.y) and \
           (rect[1] < self.y+self.height):
            return True 

        return False


    def isSelectable(self):
        return True

    def isClickable(self):
        return True

    def updateDisplay(self):
        if self.shouldUpdateColours():
            self.updateColours()

        if self.shouldUpdateImage():
            self.updateImage()
        return

    def updateImage(self):
        """
        The generic version does nothing
        """
        return

    def updateColours(self):

        colours = self.parts["colours"]
        background = self.parts[Node.PART_BACKGROUND]
        text = self.parts[Node.PART_LABEL]
        if self.isSelected():
            logging.debug("updating for selected node")
            background.borderColour = colours[NodePolicy.KEY_BORDER_SELECTION_TRUE]
            background.colour = colours[NodePolicy.KEY_BACKGROUND_SELECTION_TRUE]
            text.colour = colours[NodePolicy.KEY_TEXT_SELECTION_TRUE]
        else:
            logging.debug("updating for unselected node")
            background.borderColour = colours[NodePolicy.KEY_BORDER_SELECTION_FALSE]
            background.colour = colours[NodePolicy.KEY_BACKGROUND_SELECTION_FALSE]
            text.colour = colours[NodePolicy.KEY_TEXT_SELECTION_FALSE]
            pass

        return


    def createBackgroundPrimitive(self, size=None):
        background = PrimitivesModule.zglRect()
        background.position = [0, 0, 0]

        if size is None:
            size = self.getBounds()
        background.size = size

        background.corner_mode = True
        return background



    def getBounds(self):
        return (self.width, self.height)


    def createNamePrimitive(self, nodePolicy):
        # Fit node to text size or simply to
        # what graphviz gives us, which may not look good
        # depending on the NodePolicy
        #stringBounds = font.getSize(node.nodeData.name)
        #nodeSize = (stringBounds[0] + 2*10, stringBounds[1] + 2*10)
        nodeSize = self.getBounds()


        # initialize the name label
        nameLabel = TextModule.zglText()
        nameLabel.position = nodePolicy.getPositionForNameLabel()
        nameLabel.size = [nodeSize[0], nodeSize[1], 0]
        nameLabel.horizontal_align = nodePolicy.getHorizontalAlignmentOfLabel()
        nameLabel.vertical_align = nodePolicy.getVerticalAlignmentOfLabel()
        nameLabel.font = self.font
        nameLabel.text = self.nodeData.name()
        return nameLabel


    def setupPrimitives(self, nodePolicy):

        # self.setupDimensions()
        nodePolicy.setupDimensions(self)

        font = nodePolicy.visualPolicy().font(
            VisualPolicyModule.VisualPolicy.KEY_FONT_DEFAULT)
        self.font = font

        background = self.createBackgroundPrimitive()

        nameLabel = self.createNamePrimitive(nodePolicy)

        self.parts[Node.PART_BACKGROUND] = background
        self.parts[Node.PART_LABEL] = nameLabel
        self.parts["colours"] = nodePolicy._colours

        self.setupPorts(nodePolicy)

        self.shouldUpdateDisplay(True)

        return


    def retrieveInputPortNamesFromDataObject(self):
        if not hasattr(self.nodeData, 'inputPorts'):
            return []
        return self.nodeData.inputPorts


    def retrieveOutputPortNamesFromDataObject(self):
        if not hasattr(self.nodeData, 'outputPorts'):
            return []
        return self.nodeData.outputPorts


    def getResourceKeyForPort(self, portName):
        return 'gui port class'

    
    def setupPorts(self, nodePolicy):

        portPolicy = nodePolicy.portPolicy()

        inputPortNames = self.retrieveInputPortNamesFromDataObject()
        outputPortNames = self.retrieveOutputPortNamesFromDataObject()

        for portDirection, portNames, portMap in [
            (PortModule.Port.PORT_DIRECTION_INPUT, inputPortNames, self.inputPorts),
            (PortModule.Port.PORT_DIRECTION_OUTPUT, outputPortNames, self.outputPorts)]:

            positionCount = len(portNames)
            for index, portName in enumerate(portNames):

                resourceKey = self.getResourceKeyForPort(portName)
                portClass = self.contextManager().app().getResourceValue(
                    resourceKey, PortModule.Port)
                port = portClass(portName)
                port.node = self
                port.direction = portDirection
                port.positionIndex(index)
                port.positionCount(positionCount)

                port.initialSetup()

                portPolicy.setupDimensions(port)

                # we need to first know the port's dimensions
                # beefore we can set its position
                nodePolicy.setPortPosition(self, port)

                port.setupPrimitives(portPolicy)

                portMap[portName] = port
                pass
            pass

        return

    def getPosition(self):
        return (self.x, self.y)

    def setPosition(self, x, y):
        self.x = x
        self.y = y
        return


    def _buildChildren(self):
        return [
            self.parts[Node.PART_BACKGROUND],
            self.parts[Node.PART_LABEL]
            ] + self.inputPorts.values() + self.outputPorts.values()    


    def children(self):
        """
        A subclass can override this
        """
        if not hasattr(self, '_children'):
            # cache this so that we dont have to recompute 
            # for every redraw
            self._children = self._buildChildren()
            pass
        return self._children


    def draw(self):

        if self.shouldUpdateDisplay():
            logging.debug("updating node")
            self.updateDisplay()
            self.shouldUpdateDisplay(False)

        glPushMatrix()
        glTranslate(self.x, self.y, 0)

        for child in self.children():
            child.draw()

        glPopMatrix()

        return


    def getInputPort(self, id):
        return self.inputPorts[id]

    def getOutputPort(self, id):
        return self.outputPorts[id]

    def getSelectionContextualMenu(self, event, eventHandler, canvas, selection):

        popupMenu = SelectionContextualMenu(
            event, eventHandler, canvas, selection)
        return popupMenu


    # END class Node
    pass


class SelectionContextualMenu(MenuModule.SelectionContextualMenu):


    def bindEvents(self):

        self.addMenuItem('Edit', self.OnEdit)
        self.addMenuItem('Delete', self.OnDelete)
        return


    def OnDelete(self, event):

        # first remove all the node's incoming and outgoing edges
        objectsToRemove = []
        for edge in self.canvas.edges:
            if edge.outputNode in self.objects or \
               edge.inputNode in self.objects:
                objectsToRemove.append(edge)

        objectsToRemove.extend(self.objects)

        # remove the objects
        map(self.canvas.removeChild, objectsToRemove)

        return


    def OnEdit(self, event):
        print "%s should display editor for nodes" % self.__class
        return


    # END class SelectionContextualMenu
    pass



import math

class MouseHandler(object):
    def __init__(self, target):
        self.target = target
        self.isActive(False)
        return

    def isActive(self, value=None):
        if value is not None:
            self._isActive = value
        return self._isActive

    def updateLatestMousePosition(self, x, y):
        self.last_mouse_x = x
        self.last_mouse_y = y
        return

    def mouseButtonDown(self, button, x, y):
        self.updateLatestMousePosition(x, y)
        self.isActive(True)


    def mouseMoved(self, x, y):
        self.updateLatestMousePosition(x, y)


    def mouseButtonUp(self, button, x, y):
        self.isActive(False)


    def mouseDragged(self, x, y):
        pass



class CanvasSelectionMouseHandler (MouseHandler):
    def __init__(self, target, selection):
        MouseHandler.__init__(self,target)
        self.selection = selection
        self.is_drawing_selection_rect = False

    def mouseButtonDown(self, button, x, y):
        MouseHandler.mouseButtonDown(self, button, x, y)
        clickedObject = self.target.getObjectAtViewCoordinates(x,y)
        if clickedObject is not None:
            if not clickedObject.isSelected():
                self.selection.setSelection([clickedObject])
            return clickedObject
        else:
            self.selection.setSelection([])
            self.is_drawing_selection_rect = True
            self.selection_rect_origin = self.target.getCanvasCoordinatesFromViewCoordinates(x,y)
            return True

    def mouseDragged(self, x, y):
        if self.is_drawing_selection_rect:
            self.selection_rect_endpoint = self.target.getCanvasCoordinatesFromViewCoordinates(x,y)

            if self.hasSelectionRect():
                self.target.selection_rectangle = self.getSelectionRect()
                selectedObjects = self.target.getObjectsIntersectingRect(self.target.selection_rectangle)
                self.selection.setSelection(selectedObjects)

        self.updateLatestMousePosition(x, y)


    def mouseButtonUp(self, button, x, y):
        self.selection_rect_origin = None
        self.selection_rect_endpoint = None
        self.target.selection_rectangle = None
        MouseHandler.mouseButtonUp(self, button, x, y)

    def hasSelectionRect(self):
        return self.selection_rect_endpoint is not None and \
               self.selection_rect_origin is not None

    def getSelectionRect(self):
        if not self.hasSelectionRect():
            raise NotImplementedError('selection rect is undefined')

        x = min(self.selection_rect_endpoint[0], 
                self.selection_rect_origin[0])
        y = min(self.selection_rect_endpoint[1], 
                self.selection_rect_origin[1])
        width = max(self.selection_rect_endpoint[0], 
                    self.selection_rect_origin[0]) - x
        height = max(self.selection_rect_endpoint[1], 
                     self.selection_rect_origin[1]) - y
        return (x,y,width,height)


class CanvasScrollMouseHandler (MouseHandler):
    # For Canvas MouseHandlers, the target is the canvas itself
    def mouseDragged(self, x, y):
        delta = (x - self.last_mouse_x, y - self.last_mouse_y)
        if self.isActive():
            self.target.scroll_position[0] = self.target.scroll_position[0] - delta[0]
            self.target.scroll_position[1] = self.target.scroll_position[1] - delta[1]	
        self.updateLatestMousePosition(x, y)


class CanvasZoomMouseHandler (MouseHandler):

    def mouseButtonDown(self, button, x, y):
        MouseHandler.mouseButtonDown(self, button, x, y)
        self.mouseDownX = x
        self.mouseDownY = y
        self.originalZoom = self.target.zoom
        self.originalScroll = self.target.scroll_position[:]
        self.zoom_center = self.target.getCanvasCoordinatesFromViewCoordinates(x,y)

    def mouseDragged(self, x, y):
        delta = (x - self.last_mouse_x, y - self.last_mouse_y)
        if self.isActive() and math.fabs(delta[0]) > 0:
            delta_zoom = -0.005 * delta[0]
            newZoom = min(max(self.target.zoom + delta_zoom, 0.5), 2.0)
            self.target.zoom = newZoom

            new_zoom_center = self.target.getCanvasCoordinatesFromViewCoordinates(x, y)
            delta_scroll = (self.zoom_center[0]-new_zoom_center[0], self.zoom_center[1]-new_zoom_center[1])
            self.target.scroll_position = [self.target.scroll_position[0] + delta_scroll[0], self.target.scroll_position[1] + delta_scroll[1]]


        self.updateLatestMousePosition(x, y)


class ObjectDragMoveMouseHandler (MouseHandler):
    # Target is a list of objects being moved
    # Canvas is the canvas containing the object (needed in case of zoom)
    def __init__(self, target, canvas, delegate):
        MouseHandler.__init__(self, target)
        self.canvas = canvas
        self.delegate = delegate

    def mouseDragged(self, x, y):
        delta = (x - self.last_mouse_x, y - self.last_mouse_y)
        for object in self.target:
            object.x += delta[0] * self.canvas.zoom
            object.y += delta[1] * self.canvas.zoom

        if self.delegate is not None:
            self.delegate.objectsMoved(self.target)

        self.updateLatestMousePosition(x, y)




import os, string, sys

#print sys.path
#import Image
from PIL import Image

import math, operator		# For vector stuff

from OpenGL.GL import *
from OpenGL.GLUT import *


# Some useful constants
PI_OVER_180 = math.pi/180.0

SQRT_2 = math.sqrt(2.0)
SQRT_3 = math.sqrt(3.0)


# For conversion of row/col to matrix index, assuming OpenGL column-major 4x4 matrices
#
# utility function to get index of matrix
# row and col start at 1
#
#	0	4	8	12
#	1	5	9	13
#	2	6	10	14
#	3	7	11	15
#
#
#def rowcol(row,col):
#	return col*4 + row - 5
# prepare index
#def indexFromRowCol(row, col):
#	return row_col_index[row][col]


# A support class for vector manipulation and calculations
# There functions are all based on all vectors being an
# iterable sequence, like a list or numpy array
#
class zglVector:
	@staticmethod
	def lengthen(vector, desiredLength):
		length = len(vector)
		if length < desiredLength:
			return vector + [0]*(desiredLength-length)
		return vector[:desiredLength]
	
	@staticmethod
	def add(v1, v2):
		return map(operator.add, v1, v2)
	@staticmethod
	def subtract(v1, v2):
		return map(operator.sub, v1, v2)
	@staticmethod
	def multiply(v1, v2):
		return map(operator.mul, v1, v2)
	@staticmethod
	def divide(v1, v2):
		return map(operator.div, v1, v2)
	
	@staticmethod
	def multiplyConstant(v, c):
		return map(lambda i: i*c, v)


	@staticmethod
	def norm(v):
		return math.sqrt( reduce( lambda i,j: i + j*j, v, 0 ) )

	@staticmethod
	def normalize(v):
		v_norm = zglVector.norm(v)
		return map( lambda i: i/v_norm, v )
		
	@staticmethod
	def dotProduct(v1, v2):
		temp = map(operator.mul, v1, v2)
		return sum(temp)
	
	@staticmethod
	def crossProduct(v1, v2):
		result = [0.0, 0.0, 0.0]
		result[0] = v1[1] * v2[2] - v1[2] * v2[1];
		result[1] = v1[2] * v2[0] - v1[0] * v2[2];
		result[2] = v1[0] * v2[1] - v1[1] * v2[0];
		return result


class zglUtils:
	LEFT = 0
	CENTER = 1
	RIGHT = 2

	@staticmethod
	def pathGoUp(levels=1):
		index = path.rindex(path, delimiter)
		return path[:i], path[i+1:]
		

	@staticmethod
	def drawRect(x, y, width, height, colour):
		glPushMatrix()
		glTranslatef(x, y, 0)
		glColor4f(colour[0], colour[1], colour[2], colour[3])
		glBegin(GL_QUADS)
		glVertex2f(0, 0)
		glVertex2f(width, 0)
		glVertex2f(width, height)
		glVertex2f(0, height)
		glEnd()
		glPopMatrix()
		
	
	@staticmethod
	def drawRoundedRect( x, y, width, height, radius, colour):
		pass


	@staticmethod
	def strokeRect(x, y, width, height, colour):
		pass


	@staticmethod
	def drawBitmapText(text, x, y, align=LEFT, 
			   font=GLUT_BITMAP_HELVETICA_10, line_spacing=18):
		"""Draw the given text at given 2D position in window"""
		# Available Fonts:
		#	GLUT_BITMAP_8_BY_13
		#	GLUT_BITMAP_9_BY_15
		#	GLUT_BITMAP_TIMES_ROMAN_10
		#	GLUT_BITMAP_TIMES_ROMAN_24
		#	GLUT_BITMAP_HELVETICA_10
		#	GLUT_BITMAP_HELVETICA_12
		#	GLUT_BITMAP_HELVETICA_18
		glRasterPos2i(x, y)
		lines = 0
		for character in text:
			if character == '\n':
				glRasterPos2i(x, y-(lines*line_spacing))
			else:
				glutBitmapCharacter(font, ord(character));

	# Helper function used by multiple chart types to normalize its data
	@staticmethod
	def normalizeChartData(data, min_value, max_value):
		spread = max_value-min_value
		if spread == 0: return

		# normalize according to the min and max
		# could use list comprehension form, but some people say its slower
		# normalized_data = [float(x)/max_value for x in data]
		def normalize(x): return float(x-min_value)/spread
		normalized_data = map(normalize, data[0])
		return normalized_data

	@staticmethod
	def drawBarChart(data, x, y, width, height, colour):
		""" Draws a barchart at x,y 
		    data is assumed to be an array or tuple consisting of the following:
		    (data_array, data_min, data_max) """
		if len(data) is 0: return

		normalized_data = zglUtils.normalizeChartData(data, data[1], data[2])
		if normalized_data:
			bar_width = float(width)/float(len(normalized_data))
			xx = x
			for i in range(len(normalized_data)):
#				print "hello x=%d y=%d" % (xx, y)
				zglUtils.drawRect(xx,y, bar_width, height*normalized_data[i], colour)
				xx = xx + bar_width

	@staticmethod
	def drawBarChartWithBackground(data, x, y, width, height, colour, background_colour):
		zglUtils.drawRect(x,y,width,height,background_colour)
		glPushMatrix()
		glTranslatef(0.0, 0.0, 0.1)
		zglUtils.drawBarChart(data, x, y, width, height, colour)
		glPopMatrix()

	
	@staticmethod
	def drawLineChart(data, x, y, width, height, colour):
		""" Draws a line chart at x,y 
		    data is assumed to be an array or tuple consisting of the following:
		    (data_array, data_min, data_max) """
		if len(data) is 0: return
		elif len(data) is 1: data = data.extend(data)

		normalized_data = zglUtils.normalizeChartData(data, data[1], data[2])
		if normalized_data:
			x_increment = float(width)/float(len(normalized_data)-1)
			xx = 0
			glPushMatrix()
			glTranslatef(x,y,0)
			glColor4f(colour[0], colour[1], colour[2], colour[3])
			glBegin(GL_POLYGON)
#			glVertex2f(0,width)
#			glVertex2f(0,0)
			glVertex2f(width, 0)
			glVertex2f(0,0)
			for i in range(len(normalized_data)):
				glVertex2f(xx, height*normalized_data[i])
				xx += x_increment
			glEnd()
			glPopMatrix()
			
	
	@staticmethod
	def drawLineChartWithBackground(data, x, y, width, height, colour, background_colour):
		zglUtils.drawRect(x,y,width,height,background_colour)
		zglUtils.drawLineChart(data, x, y, width, height, colour)

	

	@staticmethod
	def getNextPowerOfTwo(num):
		""" If num isn't a power of 2, will return the next higher power of two """
		rval = 1
		while (rval<num):
			rval <<= 1
		return rval
		
	
	@staticmethod
	def loadTextureImage(path):
		# From NeHe lesson41
		
		# Load Image And Convert To A Texture
		# path can be a relative path, or a fully qualified path.
		# returns False if the requested image couldn't loaded as a texture
		# returns True and the texture ID if image was loaded
		# Catch exception here if image file couldn't be loaded
		try:
			# Note, NYI, path specified as URL's could be access using python url lib
			# OleLoadPicturePath () supports url paths, but that capability isn't critcial to this tutorial.
			Picture = Image.open(path)
		except:
			print os.getcwd()
			print "Unable to open image file '%s'." % (path)
			return 0
	
		glMaxTexDim = glGetIntegerv (GL_MAX_TEXTURE_SIZE)
	
		WidthPixels = Picture.size [0]
		HeightPixels = Picture.size [1]
	
		if ((WidthPixels > glMaxTexDim) or (HeightPixels > glMaxTexDim)):
			# The image file is too large. Shrink it to fit within the texture dimensions
			# support by our rendering context for a GL texture.
			# Note, Feel free to experiemnt and force a resize by placing a small val into
			# glMaxTexDim (e.g. 32,64,128).
			if (WidthPixels > HeightPixels):
				# Width is the domainant dimension.
				resizeWidthPixels = glMaxTexDim
				squash = float (resizeWidthPixels) / float (WidthPixels)
				resizeHeightPixels = int (HeighPixels * squash)
			else:
				resizeHeightPixels = glMaxTexDim
				squash = float (resizeHeightPixels) / float (HeightPixels)
				resizeWidthPixels = int (WidthPixels * squash)
		else:
			# // Resize Image To Closest Power Of Two
			if (WidthPixels > HeightPixels):
				# Width is the domainant dimension.
				resizeWidthPixels = zglUtils.getNextPowerOfTwo (WidthPixels)
				squash = float (resizeWidthPixels) / float (WidthPixels)
				resizeHeightPixels = int (HeightPixels * squash)
			else:
				resizeHeightPixels = zglUtils.getNextPowerOfTwo (HeightPixels)
				squash = float (resizeHeightPixels) / float (HeightPixels)
				resizeWidthPixels = int (WidthPixels * squash)
			# 
		# Resize the image to be used as a texture.
		# The Python image library provides a handy method resize (). 
		# Several filtering options are available.
		# If you don't specify a filtering option will default NEAREST
		Picture = Picture.resize((resizeWidthPixels, resizeHeightPixels), Image.BICUBIC)
		lWidthPixels = zglUtils.getNextPowerOfTwo(resizeWidthPixels)
		lHeightPixels = zglUtils.getNextPowerOfTwo(resizeWidthPixels)
		# Now we create an image that has the padding needed
		newpicture = Image.new ("RGB", (lWidthPixels, lHeightPixels), (0, 0, 0))
		
		print "picture: "+str(Picture.size[0])+","+str(Picture.size[1])+" mode="+Picture.mode
		print "new_picture: "+str(newpicture.size[0])+","+str(newpicture.size[1])+" mode="+newpicture.mode
		newpicture.paste(Picture)
	
		# Create a raw string from the image data - data will be unsigned bytes
		# RGBpad, no stride (0), and first line is top of image (-1)
		pBits = newpicture.tostring("raw", "RGBX", 0, -1)
	
		# // Typical Texture Generation Using Data From The Bitmap
		texid = glGenTextures(1);											# // Create The Texture
		glBindTexture(GL_TEXTURE_2D, texid);								# // Bind To The Texture ID
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);		# // (Modify This For The Type Of Filtering You Want)
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);     # // (Modify This For The Type Of Filtering You Want)
	
		# // (Modify This If You Want Mipmaps)
		glTexImage2D(GL_TEXTURE_2D, 0, 3, lWidthPixels, lHeightPixels, 0, GL_RGBA, GL_UNSIGNED_BYTE, pBits);
		return texid					# // Return True (All Good)





from setuptools import setup, find_packages

setup(
    name='zgl',
    version='0.1.1',
    packages=find_packages(
        'src',exclude=["*.test", "*.test.*", "test.*", "test"]),
    package_dir={'':'src'},
    test_suite="test",
    install_requires = [
        'sippy-ftgl>=0.1.0',
        'PIL',
        'numpy',
        'PyOpenGL'
        ],

    # metadata for upload to PyPI
    author = "zehao chang",
    author_email = "zehao.chang@gmail.com",
    description = "OpenGL graph drawing library",
    license = "contact developer",
    keywords = "OpenGL graph node edge dependency directed",
    url = "",
    long_description = "This package is an OpenGL based library for drawing interactive graphs",
    platforms = ["All"]

)



from zglPrimitives import *

# from http://www.python.org/workshops/1997-10/proceedings/savikko.html
class zuiEvent:
    def __init__( self, name ):
        self.name = name

class zuiWidget:
    def __init__( self, parent = None ):
        self.__parent = parent
		
    def Handle( self, event ):
        handler = 'Handle_' + event.name
        if hasattr( self, handler ):
            method = getattr( self, handler )
            method( event )
        elif self.__parent:
            self.__parent.Handle( event )
        elif hasattr( self, 'HandleDefault' ):
            self.HandleDefault( event )    

# since python supports multiple inheritance, ui objects
# is implemented simply as a collection of mouse and keybaord
# 
class zuiObject:
	def __init__(self):
		self.visible = True
		self.is_focusable = False
		self.highlight = False
		self.has_focus = False
		self.mouse_hover = False
		
	def contains(self, x, y):
		return (x >= self.position[0]) and (x <= self.position[0]+self.size[0]) and (y >= self.position[1]) and (y <= self.position[1]+self.size[1])
	
	def mousePassive(self, x, y):
		return False
	
	def mouseDown(self, button, x, y):
		if self.contains(x,y):
			self.has_focus = True
			return True
		return False

	def mouseUp(self, button, x, y):
		return False
	
	def mouseMotion(self, x, y):
		return False
		


# A simple container class for UI Objects
#
class zuiContainer(zuiObject, zglObject):
	def __init__(self):
		zglObject.__init__(self)
		zuiObject.__init__(self)
		self.elements = []
		self.current_mouse_down = None	# The object that mouse was clicked on
		self.current_focus = None	# points to object that currently has focus
		self.current_mouse_hover = None	# points to object that was last mouse hovered

	def setAlpha(self, alpha):
		self.alpha = alpha
		for element in self.elements:
			element.setAlpha(alpha)

	def setCurrentFocus(self, element):
		if self.current_focus is not None:
			self.current_focus.has_focus = False
		self.current_focus = element
		self.current_focus.has_focus = True

	def draw(self):
		self.predraw()

		if self.background is not None:
			self.background.draw()

		for element in self.elements:
			if element.visible:
				element.draw()

		self.postdraw()


	def mousePassive(self, x, y):
#		print "zuiContainer mousePassive: x=%d y=%d" % (x,y)

		if self.contains(x,y):
			# First make coordinates relative to the container origin
			x = x-self.position[0]
			y = y-self.position[1]

			self.mouse_hover = True
			for element in reversed(self.elements):
#				print "zuiContainer mousePassive checking element: x=%d y=%d" % (element.position[0],element.position[1])
				if element.mousePassive(x,y):
					# Now check if there's a previous mouse hover object
					if self.current_mouse_hover is not None and self.current_mouse_hover is not element:
						self.current_mouse_hover.mousePassive(x,y)
					self.current_mouse_hover = element
			return True
		else:
			self.mouse_hover = False
			if self.current_mouse_hover is not None:
				self.current_mouse_hover.mouse_hover = False
			return False

		
	def mouseDown(self, button, x, y):
		# First make coordinates relative to the container origin
		x = x-self.position[0]
		y = y-self.position[1]

		if not self.contains(x,y): return False
		
		self.current_mouse_down = None
		mouse_down_captured = False

		for element in reversed(self.elements):
			if element.mouseDown(button, x, y):
				if element.is_focusable:
					element.has_focus = True
					self.current_focus = element
				self.current_mouse_down = element
				mouse_down_captured = True
				break

		if not mouse_down_captured and self.current_focus is not None:
			self.current_focus.has_focus = False
			self.current_focus = None

		return True


	def mouseUp(self, button, x, y):
		# No container bounds test since you can drag an object and
		# release it outside of the container
		# Instead, only check for current_focus
		if self.current_mouse_down is not None:
			# Make coordinates relative to the container origin
			x = x-self.position[0]
			y = y-self.position[1]
			self.current_mouse_down.mouseUp(button, x, y)
			return True
		return False

	
	def mouseMotion(self, x, y):
		# First make coordinates relative to the container origin
		x = x-self.x
		y = y-self.y

		# Now check to see if any of the elements responds to
		# the mouse motion
		for element in reversed(self.elements):
			if element.mouseMotion(x, y):
				return True
		return False
		
# A simple button
#
class zuiButton(zuiObject):
	def __init__(self):
		zuiObject.__init__(self)
		# Buttons have to be focusable
		self.is_focusable = True
		
		# Flag for when mouse is pressed over the button
		self.mouse_down = False

	# Function that is called when button is clicked
	def releaseAction():
		pass
	def pressAction():
		pass
	
	def mousePassive(self, x, y):
		if self.contains(x,y):
			self.mouse_hover = True
		else:
			self.mouse_hover = False
		return self.mouse_hover
	
	def mouseDown(self, button, x, y):
		if self.contains(x,y):
			print "zuiButton mouseDown"
			self.mouse_down = True
			return True
		return False

	# The mouse action is only fired when the mouse button is released
	# on top of the button
	def mouseUp(self, button, x, y):
		print "zuiButton mouseUp"
		if self.mouse_down:
			self.mouse_down = False
			if self.contains(x,y):
				# If the mouse was released on top of the button
				self.releaseAction()
				return True
		return False
	
	def mouseMotion(self, x, y):
		if self.contains(x,y):
			self.mouse_hover = True
		else:
			self.mouse_hover = False
		return False
	


#
# slider	
class zuiSlider(zuiObject):
	def __init__(self, slider_button):
		zuiObject.__init__(self)
		
		self.slider_button = slider_button
		
		# Buttons have to be focusable
		self.is_focusable = True
		
		# Flag for when mouse is pressed over the button, ie we're dragging
		self.mouse_down = False

		
	# Function that is called when button is clicked
	def callReleaseAction(self):
		self.releaseAction()
	def releaseAction(self):
		pass
	def callPressAction(self):
		self.pressAction()
	def pressAction(self):
		pass
	def callScrollAction(self, position):
		self.scrollAction(position)
	def scrollAction(self, position):	# Position is a 0 to 1 variable of the scroller's position
		pass
		
	def mousePassive(self, x, y):
		if self.slider_button.contains(x,y):
			self.slider_button.mouse_hover = True
		else:
			self.slider_button.mouse_hover = False
		return self.slider_button.mouse_hover
	
	def mouseDown(self, button, x, y):
		if self.slider_button.contains(x,y):
			self.slider_button.mouse_down = True
			self.drag_sliderstartposition = self.slider_button.position[:]
			self.drag_startposition = (x, y)
			self.mouse_down = True
			return True
		return False

	def mouseMotion(self, x, y):
		if self.mouse_down:
			delta_x = x-self.drag_startposition[0]
			delta_y = y-self.drag_startposition[1]
			
			slider_range = self.width - self.slider_button.width
			self.slider_button.position[0] = self.drag_sliderstartposition[0] + delta_x
			
			if self.slider_button.position[0] > slider_range:
				self.slider_button.position[0] = slider_range
			elif self.slider_button.position[0] < 0:
				self.slider_button.position[0]  = 0
			
			# 0-1 variable of position
			position = self.slider_button.position[0] / slider_range
			self.callScrollAction(position)
			
		return False

	# The mouse action is only fired when the mouse button is released
	# on top of the button
	def mouseUp(self, button, x, y):
		if self.mouse_down:
			self.mouse_down = False
			if self.contains(x,y):
				# If the mouse was released on top of the button
				self.releaseAction()
				return True
		return False
	



import copy
import math

from Port import *

import zgl.zglPrimitives as PrimitivesModule
import zgl.zglText as TextModule
import zgl.zglUtils as zglUtils


# TODO:
# the following should be configurable
# port size, currently (8,8)
# min port margin, currently 5
# port spacing, currently dynamically spaced out

class PortPolicy(object):

    def __init__(self, contextManager):
        self.contextManager(contextManager)
        return

    def contextManager(self, value=None):
        if value is not None:
            self._manager = value
        return self._manager


    def setupDimensions(self, port):
        port.width = 8
        port.height = 8
        return



    # END class PortPolicy
    pass





#import currypy
import functools
import logging
import sys
import time

import wx
import numpy
import Menu as MenuModule

MODIFIERS = {
    'alt':'AltDown',
    'cmd':'CmdDown',
    'ctrl':'ControlDown',
    'meta':'MetaDown',
    'shift':'ShiftDown'
}
def getModifiers(wxEvent):
    # alt, cmd, ctrl, meta, shift
    modifiers = {}
    for key, attr in MODIFIERS.iteritems():
        modifiers[key] = getattr(wxEvent, attr)()
        pass
    return modifiers

def shouldResetMouseDownCallback(event, modifiers=None):
    if modifiers is not None:
        import Event as EventModule
        newModifiers = EventModule.getModifiers(event)
        if not newModifiers == modifiers:
            return True
    return False


class EventHandler(object):

    def eventSource(self, value=None):
        if value is not None:
            self._eventSource = value
        return self._eventSource

    def OnTimerFraction(self, event):
        # default do nothing
        return


    # END class EventHandler
    pass



class CanvasEventHandler(wx.EvtHandler, EventHandler):

    def __init__(self, eventSource):
        wx.EvtHandler.__init__(self)
        self.eventSource(eventSource)
        pass

    def bindEvents(self):
        self.Bind(wx.EVT_SIZE, self.OnSize, source=self.eventSource())
        self.Bind(wx.EVT_PAINT, self.OnPaint, source=self.eventSource())

        # call the on_timer function
        wx.EVT_TIMER(self.eventSource().GetParent(),
                     self.eventSource().TIMER_ID, 
                     self.OnTimerFraction)

        return

    def OnSize(self, evt):
        eventSource=self.eventSource()

        w,h = eventSource.GetClientSize()
        eventSource.width = w
        eventSource.height = h
        dc = wx.ClientDC(eventSource)
        eventSource.Render(dc)
        return

    def OnPaint(self, evt):
        dc = wx.PaintDC(self.eventSource())
        self.eventSource().Render(dc)
        return

    def OnTimerFraction(self, event):
        self.eventSource().forceRedraw()
        return

    # END class CanvasEventHandler
    pass


class MouseEventHandler(wx.EvtHandler, EventHandler):

    def __init__(self, eventSource):
        wx.EvtHandler.__init__(self)
        self.eventSource(eventSource)
        pass

    def bindEvents(self):
        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftMouseButtonDown, source=self.eventSource())
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftMouseButtonUpDefault, source=self.eventSource())
        self.Bind(wx.EVT_MOTION, self.OnMouseMovedDefault, source=self.eventSource())
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightMouseButtonDown, source=self.eventSource())
        self.Bind(wx.EVT_RIGHT_UP, self.OnRightMouseButtonUp, source=self.eventSource())
        return



    def computeRawMouseMove(self, initialPosition, finalPosition):

        initialMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % 
                                     (initialPosition.x, initialPosition.y))
        finalMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % 
                                   (finalPosition.x, finalPosition.y))

        return finalMatrix - initialMatrix

    def OnMouseMovedDefault(self, event):
        return




    def OnPanViewport(self, event, 
                      basePosition=None,
                      resetCallback=None,
                      shouldResetCallbackFunction=None):
        if shouldResetCallbackFunction(event):
            resetCallback()

        baseCanvasMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
                                        (basePosition.x, basePosition.y))
        baseWorldMatrix = self.eventSource().getWorldCoordinatesFromCanvasCoordinates(baseCanvasMatrix)

        newCanvasMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
                                       (event.GetX(), event.GetY()))
        newWorldMatrix = self.eventSource().getWorldCoordinatesFromCanvasCoordinates(newCanvasMatrix)

        deltaMatrix = newWorldMatrix - baseWorldMatrix

        self.eventSource().panMatrix = deltaMatrix + self.eventSource().initialPanMatrix

        return


    def OnZoomViewport(self, event, 
                       basePosition=None,
                       resetCallback=None,
                       shouldResetCallbackFunction=None):

        if shouldResetCallbackFunction(event):
            resetCallback()

        delta = self.computeRawMouseMove(basePosition, event.GetPosition())
        delta_zoom = -0.005 * delta[3,0]

        newZoomFactor = min(max(self.eventSource().initialZoomMatrix[0,0]+delta_zoom,0.5),10.0)

        newZoomMatrix = numpy.matrix('%s 0 0 0; 0 %s 0 0; 0 0 %s 0; 0 0 0 %s' %
                                     (newZoomFactor, newZoomFactor, newZoomFactor, newZoomFactor))

        self.eventSource().zoomMatrix = newZoomMatrix
        return

    def OnBoundingBoxSelection(self, event, 
                               basePosition=None,
                               resetCallback=None,
                               shouldResetCallbackFunction=None):

        baseCanvasMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
                                        (basePosition.x, basePosition.y))
        baseWorldMatrix = self.eventSource().getWorldCoordinatesFromCanvasCoordinates(baseCanvasMatrix)

        newCanvasMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' %
                                       (event.GetX(), event.GetY()))
        newWorldMatrix = self.eventSource().getWorldCoordinatesFromCanvasCoordinates(newCanvasMatrix)

        deltaMatrix = newWorldMatrix - baseWorldMatrix

        self.eventSource().selection_rectangle = (
            baseWorldMatrix[3,0], baseWorldMatrix[3,1],
            deltaMatrix[3,0], deltaMatrix[3,1]
        )


        #if self.eventSource().selection_rectangle is not None:
        #    # select the objects under the selection_rectagle
        #    pass
        objectsInBoundBox = self.eventSource().getObjectsIntersectingRect(self.eventSource().selection_rectangle)
        self.eventSource().setSelection(self.eventSource().initialSelection + objectsInBoundBox)

        return


    def OnLeftMouseButtonDown(self, event):
        point = event.GetPosition()

        # alt, cmd, ctrl, meta, shift
        modifiers = getModifiers(event)
        shouldResetCallbackFunction = functools.partial(
            shouldResetMouseDownCallback,
            modifiers=modifiers)


        mouseMoveCallback = None
        mouseUpCallback=self.OnLeftMouseButtonUpDefault

        if modifiers['alt']:
            mouseMoveCallback = self.OnPanViewport
            self.eventSource().initialPanMatrix = self.eventSource().panMatrix
            self.eventSource()
        elif modifiers['shift']:
            mouseMoveCallback = self.OnZoomViewport
            self.eventSource().initialZoomMatrix = self.eventSource().zoomMatrix
        else:
            clickedObject = self.eventSource().getClickableObjectAtCanvasCoordinates(point.x, point.y)

            if clickedObject is not None:
                # every clickable object needs to have a OnDrag callback
                mouseMoveCallback = functools.partial(
                    clickedObject.OnDrag,
                    canvas=self.eventSource())
                if clickedObject.hasCustomDragReleaseCallback():
                    mouseUpCallback = functools.partial(
                        clickedObject.OnDragRelease,
                        canvas=self.eventSource(),
                        eventHandler=self)
                    pass

                if (modifiers['ctrl'] or modifiers['cmd']) and \
                   clickedObject.isSelectable():
                    self.eventSource().addToSelection([clickedObject])
                else:
                    self.eventSource().setSelection([clickedObject])
                pass

                objects = self.eventSource().getSelection()
                mouseMoveCallback = functools.partial(
                    mouseMoveCallback,
                    objects=objects)
                for object in objects:
                    object.initialX = object.x
                    object.initialY = object.y
                pass
            else:
                if modifiers['ctrl'] or modifiers['cmd']:
                    # because the "add to" meta key is still active
                    # we don't want to reset the selection
                    pass
                else:
                    self.eventSource().setSelection([])
                self.eventSource().initialSelection = self.eventSource().getSelection()
                mouseMoveCallback = self.OnBoundingBoxSelection
                mouseUpCallback = self.resetSelectionBoundingBox
                pass


        mouseMoveCallback = functools.partial(
            mouseMoveCallback,
            basePosition=event.GetPosition(),
            resetCallback=self.resetMouseCallbacks,
            shouldResetCallbackFunction=shouldResetCallbackFunction)

        self.Bind(wx.EVT_MOTION, mouseMoveCallback, source=self.eventSource())
        self.Bind(wx.EVT_LEFT_UP, mouseUpCallback, source=self.eventSource())

        return

    def OnLeftMouseButtonUpDefault(self, event):
        self.resetMouseMovedCallback()
        return


    def resetSelectionBoundingBox(self, event):
        self.resetMouseCallbacks()
        self.eventSource().selection_rectangle=None
        return

    def resetMouseCallbacks(self):
        self.resetMouseMovedCallback()
        self.resetMouseUpCallback()
        return
    
    def resetMouseUpCallback(self):
        self.Bind(wx.EVT_MOTION, self.OnLeftMouseButtonUpDefault, source=self.eventSource())
        return

    def resetMouseMovedCallback(self):
        self.Bind(wx.EVT_MOTION, self.OnMouseMovedDefault, source=self.eventSource())
        return


    def OnRightMouseButtonDown(self, event):
        point = event.GetPosition()

        clickedObject = self.eventSource().getClickableObjectAtCanvasCoordinates(point.x, point.y)
        selection = self.eventSource().getSelection()[:]

        if clickedObject is None:
            pass
        elif len(selection) is 0 or not clickedObject in selection:
            selection = [clickedObject]

        popupMenu = None

        # create a new popup menu
        if len(selection) is 0:
            menuClass = self.eventSource().app().getResourceValue('canvas contextual menu class', MenuModule.CanvasContextualMenu)
            popupMenu = menuClass(event, self, self.eventSource())
            pass
        else:
            # get the contextual menu
            # from the most recent item selected
            popupMenu = selection[-1].getSelectionContextualMenu(
                event, self, self.eventSource(), selection)
            pass

        popupMenu.bindEvents()
        self.eventSource().PopupMenu(
            popupMenu, event.GetPosition())

        # according to http://wiki.wxpython.org/PopupMenuOnRightClick
        # need to call destroy
        popupMenu.Destroy()

        return

    def OnRightMouseButtonUp(self, event):
        point = event.GetPosition()
        print "right up at (%s,%s)" % (point.x, point.y)


        return



    # END class MouseEventHandler
    pass




class KeyEventHandler(wx.EvtHandler, EventHandler):

    def __init__(self, eventSource):
        wx.EvtHandler.__init__(self)
        self.eventSource(eventSource)
        pass

    def bindEvents(self):
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyPressed)
        self.Bind(wx.EVT_KEY_UP, self.OnKeyReleased)
        return

    def OnKeyPressed(self, event):
        modifiers = getModifiers(event)
        return

    def OnKeyReleased(self, event):
        modifiers = getModifiers(event)
        return

    # END class KeyEventHandler
    pass


class MyPopupMenu(wx.Menu):
    def __init__(self, parent):
        wx.Menu.__init__(self)

        self.parent = parent

        minimize = wx.MenuItem(self, wx.NewId(), 'Minimize')
        self.AppendItem(minimize)
        self.Bind(wx.EVT_MENU, self.OnMinimize, id=minimize.GetId())

        close = wx.MenuItem(self, wx.NewId(), 'Close')
        self.AppendItem(close)
        self.Bind(wx.EVT_MENU, self.OnClose, id=close.GetId())


    def OnMinimize(self, event):
        self.parent.Iconize()

    def OnClose(self, event):
        self.parent.Close()





import logging
import sys
import time

from zgl_graphdrawer.Edge import *
from zgl_graphdrawer.Node import *
from zgl_graphdrawer.MouseHandlers import *

import zgl_graphdrawer.canvas as CanvasModule
import zgl_graphdrawer.Event as EventModule
import zgl_graphdrawer.GraphDrawerManager as ContextManagerModule

from zgl.zglUtils import zglUtils

import OpenGL
from OpenGL.GL import *

import numpy

class Point(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y
        return

class Canvas(CanvasModule.Canvas):

    def __init__(self, parent, width, height, *args, **kwds):

        CanvasModule.Canvas.__init__(self)

        self.width = width
        self.height = height
        self.initGL()
        self.initLighting()
        self.initViewport()

        self._selected = []

        return

    def resetDrawables(self):
        CanvasModule.Canvas.resetDrawables(self)
        self.resetDataToGuiObjectMapping()
        return

    def resetDataToGuiObjectMapping(self):
        self._dataToGuiObjectMapping = {}
        return

    def addDataToGuiObjectMapping(self, dataObject, guiObject):
        self._dataToGuiObjectMapping[dataObject] = guiObject
        return

    def getGuiObjectForDataObject(self, dataObject):
        return self._dataToGuiObjectMapping[dataObject]
 

    def initializeEventHandlers(self, contextManager=None):
        self.contextManager(contextManager)

        for key, default in [
            ('mouse event handler', EventModule.MouseEventHandler),
            ('key event handler', EventModule.KeyEventHandler),
            ('canvas event handler', EventModule.CanvasEventHandler),
            ]:

            classObject = contextManager.app().getResourceValue(
                "%s class" % key, default=default)

            eventHandler = classObject(self)
            eventHandler.bindEvents()
            self.PushEventHandler(eventHandler)

            # TODO:
            # remove this once the event handler mechanism works
            contextManager.app().setResourceValue(
                key, eventHandler)

            pass

        return

    def getHeight(self):
        return self.height

    def getWidth(self):
        return self.width

    def resize(self, width, height):
        self.width = width
        self.height = height
        self.setViewport(width, height)
        return
    
    def initViewport(self):
        return self._initViewport(self.getWidth(), self.getHeight())

    def _initViewport(self, width, height):

        self.setViewport(width, height)
        self.panMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; 0 0 0 0')

        return

    def setViewport(self, width, height):
        # Reset The Current Viewport And Perspective Transformation
        glViewport(0, 0, width, height)		
        self._updateProjection(width, height)
        return

    def initLighting(self):
        return self._initLighting()

    def _initLighting(self):
        return

    def initGL(self):
        return self._initGL()

    def _initGL(self):

        glClearDepth (1.0)
        glEnable (GL_DEPTH_TEST)
        # glClearColor (0.0, 0.0, 0.0, 0.0)
        glClearColor (0.6, 0.6, 0.6, 1.0)
        
        glShadeModel (GL_SMOOTH)
        glMatrixMode (GL_PROJECTION)

        # Enable blending
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        # Really Nice Perspective Calculations
        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST) 

        # Enables Smooth Color Shading
        glShadeModel(GL_SMOOTH)
        # Enables Clearing Of The Depth Buffer
        glClearDepth(1.0)

        # Enables Depth Testing
        glEnable(GL_DEPTH_TEST)

        # The Type Of Depth Test To Do	
        glDepthFunc(GL_LEQUAL)

        glDisable(GL_LIGHTING)
        glColor3f(1.0, 1.0, 1.0)

        self.updateProjection()

        return


    def Render(self, dc):
        self.SetCurrent()

        self.displayFunc()

        self.SwapBuffers()
        return

    def displayFunc(self):
        return self._displayFunc()

    def _displayFunc(self):
        glDrawBuffer(GL_BACK)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        glDisable(GL_LIGHTING)
        glColor3f(1.0, 1.0, 1.0)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        """main drawing stuff goes here"""
        self.draw()

        self.drawMetaInfo()
        return


    def drawMetaInfo(self):

        x = 30
        y = self.getHeight() - 30
        self.drawFPS(x, y)

        return


    def updateProjection(self):
        return self._updateProjection(self.getWidth(), self.getHeight())


    def _updateProjection(self, width, height):
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(0.0, width, 0.0, height)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        return


    def drawFPS(self, x, y):
        if hasattr(self, "last_t"):
            self.fps = 1.0/float(time.time()-self.last_t)
            glColor3f(1.0, 1.0, 1.0)

            zglUtils.drawBitmapText(
                "FPS: %f" % (self.fps), 
                x, y,
                font=GLUT_BITMAP_HELVETICA_10)
        self.last_t = time.time()

        return



    def getClickableObjectAtCanvasCoordinates(self, x, y):
        canvasMatrix = numpy.matrix('0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % (x, y))
        worldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(canvasMatrix)
        world_x = worldMatrix[3,0]
        world_y = worldMatrix[3,1]

        point = Point(world_x, world_y)

        for clickable in self.getClickables():
            if clickable.overlaps(point):
                return clickable

        return None

    def getSelectableObjectAtCanvasCoordinates(self, x, y):

        """
        this should instead be a composite filter
        # object is selectable
        # object contains point
        """

        canvasMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; %s %s 0 0' % (x, y))
        worldMatrix = self.getWorldCoordinatesFromCanvasCoordinates(canvasMatrix)
        world_x = worldMatrix[3,0]
        world_y = worldMatrix[3,1]

        point = Point(world_x, world_y)

        for selectable in self.getSelectables():
            if selectable.overlaps(point):
                return selectable

        return None

    def getViewportCoordinatesFromCanvasCoordinates(self, canvasMatrix):
        rotationMatrix = numpy.matrix(
            '1 0 0 0; 0 -1 0 0; 0 0 -1 0; 0 0 0 1')
        translationMatrix = numpy.matrix(
            '0 0 0 0; 0 0 0 0; 0 0 0 0; 0 %s 0 0' % self.getHeight())

        viewportMatrix = (canvasMatrix*rotationMatrix) +translationMatrix
        return viewportMatrix


    def getWorldCoordinatesFromViewportCoordinates(self, viewportMatrix):

        worldMatrix = self.zoomMatrix * (viewportMatrix - self.panMatrix)

        return worldMatrix


    def getWorldCoordinatesFromCanvasCoordinates(self, canvasMatrix):
        viewportMatrix = self.getViewportCoordinatesFromCanvasCoordinates(canvasMatrix)
        worldMatrix = self.getWorldCoordinatesFromViewportCoordinates(viewportMatrix)

        return worldMatrix


    def setSelection(self, selected):

        oldSelection = self._selected
        self._selected = selected

        self.contextManager().selectionChanged(oldSelection, self._selected)

        return

    def addToSelection(self, selected):

        self._selected.extend(selected)

        self.contextManager().selectionChanged([], self._selected)
        pass


    def getSelection(self):
        return self._selected


    def getResourceKeyForNode(self, dataNode):
        return 'gui node class'


    def addNode(self, dataNode, worldMatrix=None):

        resourceKey = self.getResourceKeyForNode(dataNode)
        nodeClass = self.contextManager().app().getResourceValue(
            resourceKey, default=Node)
        
        uiNode = nodeClass(dataNode)
        # need to do this so that the node
        # will have access to app contextual information
        uiNode.contextManager(self.contextManager())
        
        uiNode.setupPrimitives(self.nodePolicy())

        self.addDrawable(uiNode)

        if worldMatrix is None:
            worldMatrix = numpy.matrix('1 0 0 0; 0 1 0 0; 0 0 1 0; 0 0 0 1')

        uiNode.setPosition(worldMatrix[3,0], worldMatrix[3,1])

        dataObjectFunction = self.addDataToGuiObjectMapping

        dataObjectFunction(dataNode, uiNode)
        ports = uiNode.inputPorts.values()+uiNode.outputPorts.values()
        for port in ports:
            portName = port.name

            # parameters have __eq__ defined
            # such that they are equivalent if their ids are equivalent
            # and since the ids are not unique (and are contextualized)
            # we need to include the context, ie the data node
            dataObjectFunction((dataNode, portName), port)

            pass

        return uiNode


    def addEdge(self, dataPath):
        # find the gui objects
        # for the outputNode, outputPort, inputPort, inputNode

        dataObjectFunction = self.getGuiObjectForDataObject

        guiPath = map(
            dataObjectFunction,
            [
                dataPath[0], # source node
                (dataPath[0], dataPath[1]), # source port
                (dataPath[-1], dataPath[-2]), # target port
                dataPath[-1] # target node
            ]
        )

        edgeClass = self.contextManager().app().getResourceValue(
            'gui edge class', default=Edge)

        edge = edgeClass(dataPath, guiPath)
        edge.setupPrimitives(self.edgePolicy())
        self.addDrawable(edge)

        return


    # END class Canvas
    pass

